# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""The eqdsk package provides a class for reading and writing eqdsk files.

Module contents:

- [`file.py`][eqdsk.file]: The eqdsk file class.
- `cocos`: The eqdsk coordinate system conversion functions.

"""

from eqdsk.file import EQDSKInterface

__all__ = ["EQDSKInterface"]
